@@ -6,14 +6,18 @@ import subprocess
 import logging
 import json
 import re
+import secrets
+import cairo
+from pathlib import Path
 
-from xapp.GSettingsWidgets import GSettingsSwitch, GSettingsFileChooser, GSettingsComboBox
+from xapp.GSettingsWidgets import GSettingsSwitch, GSettingsFileChooser, GSettingsComboBox, GSettingsSpinButton
 from xapp.SettingsWidgets import SettingsWidget, SettingsPage, SettingsStack, SpinButton, Entry, Button, ComboBox
-from gi.repository import Gtk, Gio, GLib
+from gi.repository import Gtk, Gdk, Gio, GLib
 
 import config
 import auth
 import util
+import misc
 import networkmonitor
 
 _ = gettext.gettext
@@ -38,24 +42,82 @@ RPC_THREAD_POOL_SIZE_KEY = "rpc-thread-pool-size"
 USE_COMPRESSION_KEY = "use-compression"
 COMPRESSION_LEVEL_KEY = "zlib-compression-level"
 BLOCK_SIZE_KEY = "transfer-block-size"
+MIN_FREE_SPACE_KEY = "minimum-free-space"
+GROUP_CODE_KEY = "group-code"
+CONNECT_ID_KEY = "connect-id"
+
+DEFAULT_GROUP_CODE = "Warpinator"
 
 prefs_settings = Gio.Settings(schema_id=PREFS_SCHEMA)
 
-#### One-time initializers
-if prefs_settings.get_string(FOLDER_NAME_KEY) == "":
-    default = Gio.File.new_for_path(os.path.join(GLib.get_home_dir(), "Warpinator"))
+## Migrate ~/.config/warpinator/.group
+
+KEYFILE_GROUP_NAME = "warpinator"
+KEYFILE_CODE_KEY = "code"
+KEYFILE_UUID_KEY = "connect_id"
+CONFIG_FILE_NAME = ".group"
+CONFIG_FOLDER = Path(os.path.join(GLib.get_user_config_dir(), "warpinator"))
+path = Path(os.path.join(CONFIG_FOLDER, CONFIG_FILE_NAME))
+
+def get_new_connect_id():
+    return "%s-%s" % (util.get_hostname().upper()[:42], secrets.token_hex(10).upper())
+
+try:
+    keyfile = GLib.KeyFile()
+    keyfile.load_from_file(path.as_posix(), GLib.KeyFileFlags.NONE)
 
     try:
-        os.makedirs(default.get_path(), exist_ok=True)
+        code = keyfile.get_string(KEYFILE_GROUP_NAME, KEYFILE_CODE_KEY)
+
+        if code == None or code == "":
+            raise
+
+        if len(code) < 4:
+            logging.warn("Group Code is short, consider something longer than 8 characters.")
     except:
-        default = Gio.File.new_for_path(GLib.get_home_dir())
+        code = DEFAULT_GROUP_CODE
 
-    prefs_settings.set_string(FOLDER_NAME_KEY, default.get_uri())
-####
+    try:
+        connect_id = keyfile.get_string(KEYFILE_GROUP_NAME, KEYFILE_UUID_KEY)
+        if len(connect_id.split("-")) == 5:
+            raise
+    except:
+        # Max 'instance' length is 63.
+        # https://datatracker.ietf.org/doc/html/rfc6763#section-7.2
+        connect_id = get_new_connect_id()
 
+    prefs_settings.set_string(GROUP_CODE_KEY, code)
+    prefs_settings.set_string(CONNECT_ID_KEY, connect_id)
+
+    path.unlink()
+
+    try:
+        path.parent.rmdir()
+    except (OSError, FileNotFoundError):
+        logging.warn("Could not remove obsolete group code file and directory at '%s' - maybe the directory isn't empty?" % str(path.parent))
+except GLib.Error as e:
+    logging.debug("Migration failed - either migration already happened, or there was nothing to migrate in the first place: %s" % str(e))
+
+## /migrate
+
+# Sanity checks, initial values...
 if prefs_settings.get_int(PORT_KEY) == prefs_settings.get_int(REG_PORT_KEY):
     prefs_settings.set_int(REG_PORT_KEY, prefs_settings.get_int(PORT_KEY) + 1)
 
+code = prefs_settings.get_string(GROUP_CODE_KEY)
+if code == "":
+    prefs_settings.set_string(GROUP_CODE_KEY, DEFAULT_GROUP_CODE)
+connect_id = prefs_settings.get_string(CONNECT_ID_KEY)
+if connect_id == "":
+    prefs_settings.set_string(CONNECT_ID_KEY, get_new_connect_id())
+
+prefs_settings.sync()
+
+# /sanity
+
+def get_should_autostart():
+    return prefs_settings.get_boolean(AUTOSTART_KEY)
+
 def get_preferred_iface():
     return prefs_settings.get_string(NET_IFACE)
 
@@ -66,6 +128,25 @@ def get_auth_port():
     return prefs_settings.get_int(REG_PORT_KEY)
 
 def get_save_uri():
+    if prefs_settings.get_string(FOLDER_NAME_KEY) == "":
+        logging.info("No save location set")
+        parent = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOWNLOAD)
+
+        if not parent:
+            parent = GLib.get_home_dir()
+
+        default = Gio.File.new_for_path(os.path.join(parent, "Warpinator"))
+
+        try:
+            os.makedirs(default.get_path(), exist_ok=True)
+            logging.info("Created default save directory: '%s'" % default.get_path())
+        except:
+            default = Gio.File.new_for_path(GLib.get_home_dir())
+            logging.warning("Could not create default save directory - using '%s'" % default.get_path())
+
+        prefs_settings.set_string(FOLDER_NAME_KEY, default.get_uri())
+        prefs_settings.sync()
+
     uri = prefs_settings.get_string(FOLDER_NAME_KEY)
 
     return uri
@@ -113,6 +194,9 @@ def get_remote_pool_max_threads():
     if setting_value == 0:
         setting_value = max(8, os.cpu_count() + 4)
 
+    if setting_value < 4:
+        setting_value = 4
+
     return setting_value
 
 def get_server_pool_max_threads():
@@ -134,8 +218,55 @@ def get_compression_level():
 def get_block_size():
     return prefs_settings.get_int(BLOCK_SIZE_KEY) * 1024
 
+def get_min_free_space():
+    return prefs_settings.get_uint(MIN_FREE_SPACE_KEY)
+
+def get_group_code():
+    return prefs_settings.get_string(GROUP_CODE_KEY)
+
+def get_secure_mode():
+    return get_group_code() != DEFAULT_GROUP_CODE
+
+def get_connect_id():
+    return prefs_settings.get_string(CONNECT_ID_KEY)
+
+
+#### Secure mode
+class SecureModePrefsBlocker():
+    # This prevents external changes to Warpinator (like from a terminal or dconf-editor) while warpinator
+    # is running
+    def __init__(self):
+        self.active = False
+        self.blocker_settings = Gio.Settings(schema_id=PREFS_SCHEMA)
+        self.blocker_settings.delay()
+
+        self.settings_changed_id = 0
+
+        self._enforce_settings()
+
+    def _enforce_settings(self, settings=None, key=None):
+        if get_group_code() != DEFAULT_GROUP_CODE:
+            return
+
+        if self.settings_changed_id > 0:
+            self.blocker_settings.handler_block(self.settings_changed_id)
+
+        self.blocker_settings.set_boolean(AUTOSTART_KEY, False)
+        self.blocker_settings.set_boolean(ASK_PERMISSION_KEY, True)
+        self.blocker_settings.set_boolean(NO_OVERWRITE_KEY, True)
+        self.blocker_settings.apply()
+        self.blocker_settings.sync() # when this is used in /usr/bin/warpinator to check for autostart, there's no main loop yet.
+
+        if self.settings_changed_id > 0:
+            self.blocker_settings.handler_unblock(self.settings_changed_id)
+
+    def start_monitor(self):
+        self.settings_changed_id = self.blocker_settings.connect("changed", self._enforce_settings)
+
+####
+
 class Preferences():
-    def __init__(self, transient_for):
+    def __init__(self, main_window, page_name):
         self.builder = Gtk.Builder.new_from_file(os.path.join(config.pkgdatadir, "prefs-window.ui"))
 
         self.window = self.builder.get_object("prefs_window")
@@ -145,7 +276,7 @@ class Preferences():
         self.window.set_title(title=_("Warpinator Preferences"))
         self.window.set_icon_name("preferences-system")
 
-        self.window.set_transient_for(transient_for)
+        self.window.set_transient_for(main_window)
 
         size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.HORIZONTAL)
 
@@ -153,6 +284,10 @@ class Preferences():
         self.content_box.pack_start(page_stack, True, True, 0)
         self.page_switcher.set_stack(page_stack)
 
+        prefs_settings.connect("changed::group-code", self.on_group_code_changed)
+
+        self.unsafe_options = []
+
         # Settings
         page = SettingsPage()
         page_stack.add_titled(page, "general", _("General"))
@@ -170,6 +305,7 @@ class Preferences():
 
         widget = GSettingsSwitch(_("Start automatically"),
                                  PREFS_SCHEMA, AUTOSTART_KEY)
+        self.unsafe_options.append(widget)
         section.add_row(widget)
 
         section = page.add_section(_("File Transfers"))
@@ -184,12 +320,19 @@ class Preferences():
                                       size_group=size_group, dir_select=True)
         section.add_row(widget)
 
+        widget = GSettingsSpinButton(_("Reserved free space"),
+                                     PREFS_SCHEMA, MIN_FREE_SPACE_KEY, units="MB", mini=250, maxi=GLib.MAXUINT)
+
+        section.add_row(widget)
+
         widget = GSettingsSwitch(_("Require approval before accepting files"),
                                  PREFS_SCHEMA, ASK_PERMISSION_KEY)
+        self.unsafe_options.append(widget)
         section.add_row(widget)
 
         widget = GSettingsSwitch(_("Require approval when files would be overwritten"),
                                  PREFS_SCHEMA, NO_OVERWRITE_KEY)
+        self.unsafe_options.append(widget)
         section.add_row(widget)
 
         widget = GSettingsSwitch(_("Display a notification when someone sends (or tries to send) you files"),
@@ -199,16 +342,12 @@ class Preferences():
         page = SettingsPage()
         page_stack.add_titled(page, "network", _("Connection"))
 
-        section = page.add_section(_("Identification"))
+        section = page.add_section(_("Group Code"))
 
         entry_size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.VERTICAL)
         button_size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.BOTH)
 
-        widget = GroupCodeEntry(_("Group Code"),
-                                tooltip=_("You cannot communicate with computers that do not use the same code."),
-                                entry_size_group=entry_size_group,
-                                button_size_group=button_size_group)
-
+        widget = GroupCodeEntry(focus_entry = page_name == "network")
         section.add_row(widget)
 
         section = page.add_section(_("Network"))
@@ -226,9 +365,9 @@ class Preferences():
                                                stderr=subprocess.DEVNULL
                                               ).decode("utf-8")
 
-            lshw_out = re.sub("\A\s*{", "[{", lshw_out)
-            lshw_out = re.sub("}\s*{", "}, {", lshw_out)
-            lshw_out = re.sub("}\s*\\n\Z", "}]\n", lshw_out)
+            lshw_out = re.sub(r"\A\s*{", "[{", lshw_out)
+            lshw_out = re.sub(r"}\s*{", "}, {", lshw_out)
+            lshw_out = re.sub(r"}\s*\\n\Z", "}]\n", lshw_out)
             j = json.loads(lshw_out)
         except:
             pass
@@ -280,7 +419,7 @@ class Preferences():
             if iface == current:
                 current_selection_exists = True
 
-            if (desc != None and desc != ""):
+            if desc is not None and desc != "":
                 orig_label = "%s - %s" % (iface, desc)
                 if len(orig_label) > 50:
                     label = orig_label[:47] + "..."
@@ -353,7 +492,13 @@ can make it simpler to add firewall exceptions if necessary."""))
 
         section.add_row(widget)
 
+        self.on_group_code_changed(None, None)
+
         self.window.show_all()
+        page_stack.set_visible_child_full(page_name, transition=Gtk.StackTransitionType.NONE)
+
+    def destroy(self):
+        GLib.timeout_add(1000, self.window.destroy)
 
     def open_port(self, widget):
         self.run_port_script(self.settings.get_int(PORT_KEY), self.settings.get_int(REG_PORT_KEY))
@@ -364,7 +509,7 @@ can make it simpler to add firewall exceptions if necessary."""))
 
         iface_widget_value = None
         tree_iter = self.iface_combo.content_widget.get_active_iter()
-        if tree_iter != None:
+        if tree_iter is not None:
             iface_widget_value = self.iface_combo.model[tree_iter][0]
 
         if main_widget_value == get_port() and \
@@ -387,10 +532,10 @@ can make it simpler to add firewall exceptions if necessary."""))
 
         iface_widget_value = None
         tree_iter = self.iface_combo.content_widget.get_active_iter()
-        if tree_iter != None:
+        if tree_iter is not None:
             iface_widget_value = self.iface_combo.model[tree_iter][0]
 
-        if iface_widget_value == None:
+        if iface_widget_value is None:
             iface_widget_value = "auto"
 
             self.iface_combo.content_widget.disconnect(self.iface_combo_id)
@@ -403,13 +548,24 @@ can make it simpler to add firewall exceptions if necessary."""))
 
         self.settings.apply()
 
-    @util._async
+    @misc._async
     def run_port_script(self, port, auth_port):
         command = os.path.join(config.libexecdir, "firewall", "ufw-modify")
         subprocess.run(["pkexec", command, str(port), str(auth_port)])
 
         GLib.timeout_add_seconds(1, lambda: Gio.Application.get_default().firewall_script_finished())
 
+    def on_group_code_changed(self, settings, key):
+        is_default_code = not get_secure_mode()
+
+        for widget in self.unsafe_options:
+            if is_default_code:
+                widget.set_sensitive(False)
+                widget.set_tooltip_text(_("Only available in secure mode."))
+            else:
+                widget.set_sensitive(True)
+                widget.set_tooltip_text(None)
+
 class PortSpinButton(SpinButton):
     def __init__(self, *args, **kargs):
         button_size_group = kargs.pop("button_size_group")
@@ -431,52 +587,98 @@ class PortSpinButton(SpinButton):
     def apply_later(self, *args):
         pass
 
-class GroupCodeEntry(Entry):
-    def __init__(self, *args, **kargs):
-        button_size_group = kargs.pop("button_size_group")
-        entry_size_group = kargs.pop("entry_size_group")
+class GroupCodeEntry(SettingsWidget):
+    def __init__(self, focus_entry=False):
+        super(GroupCodeEntry, self).__init__()
 
-        super(GroupCodeEntry, self).__init__(*args, **kargs)
+        self.code = get_group_code()
+        prefs_settings.connect("changed::group-code", self.on_group_code_changed)
 
-        self.code = auth.get_singleton().get_group_code()
-        self.content_widget.set_text(self.code)
+        self.builder = Gtk.Builder.new_from_file(os.path.join(config.pkgdatadir, "group-code.ui"))
 
-        entry_size_group.add_widget(self.content_widget)
-        self.content_widget.connect("changed", self.text_changed)
+        self.toplevel = self.builder.get_object("toplevel")
+        self.pack_start(self.toplevel, True, True, 0)
 
-        self.set_child_packing(self.content_widget, False, False, 0, Gtk.PackType.END)
-        self.set_spacing(6)
+        self.entry = self.builder.get_object("code_entry")
 
-        self.accept_button = Gtk.Button(label=_("Set code"))
-        self.accept_button.show()
-        self.accept_button.set_sensitive(False)
-        self.accept_button.get_style_context().add_class("suggested-action")
-        self.accept_button.connect("clicked", self.apply_clicked)
-        button_size_group.add_widget(self.accept_button)
+        if focus_entry:
+            self.entry.connect("realize", lambda w: w.grab_focus())
 
-        self.pack_end(self.accept_button, False, False, 0)
-        self.reorder_child(self.accept_button, 1)
+        self.set_code_button = self.builder.get_object("set_code_button")
+        self.more_info_link_button = self.builder.get_object("more_info_link_button")
+        self.secure_mode_label = self.builder.get_object("secure_mode_label")
+        self.reason_label = self.builder.get_object("reason_label")
+        self.status_bar = self.builder.get_object("status_bar")
+
+        self.entry.set_text(self.code)
+        self.entry.connect("changed", self.text_changed)
+
+        self.set_code_button.set_sensitive(False)
+        self.set_code_button.get_style_context().add_class("suggested-action")
+        self.set_code_button.connect("clicked", self.set_code_clicked)
+
+        context = self.get_style_context()
+
+        ret, self.secure_color = context.lookup_color("success_color")
+        if not ret:
+            self.secure_color = Gdk.RGBA()
+            self.secure_color.parse("green")
+
+        ret, self.insecure_color = context.lookup_color("error_color")
+        if not ret:
+            self.insecure_color = Gdk.RGBA()
+            self.insecure_color.parse("red")
+
+        self.status_bar.connect("draw", self.status_bar_draw)
+
+        self.on_group_code_changed(None, None)
 
     def text_changed(self, widget, data=None):
-        text = self.content_widget.get_text()
+        text = widget.get_text()
 
         if text == "":
-            widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "dialog-error-symbolic")
+            widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "xsi-dialog-error-symbolic")
             widget.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("A group code is required."))
-            self.accept_button.set_sensitive(False)
+            self.set_code_button.set_sensitive(False)
+            return
+        elif len(text) < 4:
+            widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "xsi-dialog-error-symbolic")
+            widget.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("The group code is too short."))
+            self.set_code_button.set_sensitive(False)
             return
-
-        if len(text) < 8:
-            widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "dialog-warning-symbolic")
-            widget.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, _("The group code should be longer if possible."))
         else:
             widget.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, None)
 
-        self.accept_button.set_sensitive(self.content_widget.get_text() != self.code)
+        self.set_code_button.set_sensitive(widget.get_text() != self.code)
 
-    def apply_clicked(self, widget, data=None):
-        self.code = self.content_widget.get_text()
-        self.accept_button.set_sensitive(False)
-        auth.get_singleton().update_group_code(self.code)
+    def set_code_clicked(self, widget, data=None):
+        self.code = self.entry.get_text()
+        self.set_code_button.set_sensitive(False)
+        prefs_settings.set_string(GROUP_CODE_KEY, self.code)
 
+    def status_bar_draw(self, widget, cr):
+        if get_secure_mode():
+            color = self.secure_color
+        else:
+            color = self.insecure_color
 
+        allocation = self.get_allocation()
+        cr.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
+
+        cr.save()
+
+        Gdk.cairo_set_source_rgba(cr, color)
+        cr.rectangle(0, 0, allocation.width, allocation.height)
+        cr.fill()
+
+        cr.restore()
+
+        return True
+
+    def on_group_code_changed(self, settings, key):
+        if get_secure_mode():
+            self.secure_mode_label.set_text(_("ON"))
+            self.reason_label.set_text(_("All options are unlocked."))
+        else:
+            self.secure_mode_label.set_text(_("OFF"))
+            self.reason_label.set_text(_("You must set a custom group code to run in secure mode. Some options are currently disabled."))
